import heapq
import re
from collections import namedtuple
from typing import List, Tuple

from avalares.pattern_detector import PatternDetector, PatternMarking

token_matcher = r'(?:{})'.format('|'.join([
    r'(?P<string>[a-zA-Z_][a-zA-Z_0-9]*)',
    r'(?P<float>[0-9]+\.[0-9]+)',
    r'(?P<int>[0-9]+)',
    r'(?P<space> +)'
]))

value_types = {
    'string', 'float', 'int'
}

converters = {
    'string': lambda x: x,
    'float': float,
    'int': int
}

ParseResult = namedtuple('ParseResult', 'rows labels types')
TokenData = namedtuple('TokenData', 'labels values')


def parse(text: str, convert_values=True) -> ParseResult:
    data = _tokenize_string(text.strip())
    detector = PatternDetector()
    for width in range(2, len(data.labels) + 1):
        for i in range(width, len(data.labels) + 2, width):
            vals = data.labels[i - width:i]
            if len(vals) < width:
                prev = data.labels[i - 2 * width:i - width]
                vals.extend(prev[len(vals):])
            detector.mark_pattern(vals, i - width)
        detector.finish()
    score = lambda x: x.count ** 2 * len(x.pattern)
    marking = heapq.nlargest(1, detector.pattern_counts, key=score)[0]

    delim = _detect_delimiter(data.labels[marking.start_pos:], marking.pattern)
    data_start, schema = _fix_offset(data, marking, delim)
    rows = _extract_rows(data, data_start, schema, delim, convert_values)
    label_names = _try_extract_header(data, data_start, delim, schema)
    types = [i for i in marking.pattern if i in value_types]
    return ParseResult(rows, label_names, types)


def _extract_rows(data: TokenData, pos: int, schema: tuple, delim: str, convert_values=True):
    rows = []
    while data.labels[pos:pos + len(schema)] == schema:
        labels = data.labels[pos:pos + len(schema)]
        values = data.values[pos:pos + len(schema)]
        rows.append(tuple(
            converters[label](value) if convert_values else value
            for label, value in zip(labels, values)
            if label in value_types
        ))
        pos += len(schema)
        if pos >= len(data.labels) or data.labels[pos] != delim:
            break
        pos += 1
    return rows


def _fix_offset(data: TokenData, marking: PatternMarking, delim: str) -> Tuple[int, tuple]:
    pattern = marking.pattern
    if delim not in pattern:
        return marking.start_pos, pattern
    delim_pos = pattern.index(delim)
    schema = pattern[delim_pos + 1:] + pattern[:delim_pos]
    pos = marking.start_pos - len(pattern[delim_pos + 1:])
    if data.labels[pos:pos + len(schema)] != schema:
        pos += len(schema) + 1
    data_start = pos
    return data_start, schema


def _detect_delimiter(data_labels: list, pattern: tuple) -> str:
    chars = set(data_labels)
    last_char = {c: -1 for c in chars}
    detectors = {c: PatternDetector() for c in chars}
    for i, c in enumerate(data_labels):
        dist = i - last_char[c]
        detectors[c].mark_pattern(dist, last_char[c])
        last_char[c] = i
    for c, last_i in last_char.items():
        dist = len(data_labels) - last_i
        detectors[c].mark_pattern(dist, len(data_labels))
        detectors[c].finish()

    return max(detectors, key=lambda x: max(
        (count ** 2 * interval if interval >= len(pattern) else 0, x)
        for count, start_pos, interval in detectors[x].pattern_counts
    ))


def _tokenize_string(text: str) -> TokenData:
    token_labels = []
    token_values = []
    next_char = 0
    for match in re.finditer(token_matcher, text):
        a, b = match.span()
        if a > next_char:
            token_labels.extend(text[next_char:a])
            token_values.extend(text[next_char:a])
        label, value = next(iter(((k, v) for k, v in match.groupdict().items() if v is not None)))
        token_labels.append(label)
        token_values.append(value)
        next_char = b
    a = len(text)
    if a > next_char:
        token_labels.extend(text[next_char:a])
        token_values.extend(text[next_char:a])
    return TokenData(token_labels, token_values)


def _step_back_line(tokens: TokenData, pos: int, delim: str):
    end_pos = pos - 1
    pos = pos - 2
    while pos >= 0 and tokens.labels[pos] != delim:
        pos -= 1
    pos += 1
    return pos, end_pos


def _try_extract_header(tokens: TokenData, data_start: int, delim: str, schema: tuple) -> List[str]:
    label_names = []
    pos, end_pos = _step_back_line(tokens, data_start, delim)
    while pos >= 0:
        num_schema_vals = sum(i in value_types for i in schema)
        num_label_vals = sum(i in value_types for i in tokens.labels[pos:end_pos])
        if num_label_vals == 0:
            pos, end_pos = _step_back_line(tokens, pos, delim)
            continue
        if num_schema_vals == num_label_vals:
            for token, value in zip(tokens.labels[pos:end_pos], tokens.values[pos:end_pos]):
                if token in value_types:
                    label_names.append(value)
        break
    return label_names



from collections import namedtuple

from typing import List

PatternMarking = namedtuple('PatternMarking', 'count start_pos pattern')


class PatternDetector:
    def __init__(self):
        self.last_pattern = None
        self.count = 0
        self.start_pos = None
        self.pattern_counts = []  # type: List[PatternMarking]

    def mark_pattern(self, pattern, start_pos):
        if pattern != self.last_pattern:
            self.finish()
            self.count = 1
            self.start_pos = start_pos
            self.last_pattern = pattern
        else:
            self.count += 1

    def finish(self):
        if self.last_pattern is not None:
            self.pattern_counts.append(PatternMarking(self.count, self.start_pos, self.last_pattern))



"""
Tool to automatically extract a pandas dataframe or numpy array out of a string of text
"""
import csv
import io
import json
import pickle
import sys
from argparse import ArgumentParser
from os.path import splitext

from avalares.parser import parse


def to_json(data, label_names=None):
    return [
        {label_name: value for label_name, value in zip(label_names, row)}
        if label_names else
        list(row)
        for row in data
    ]


def write_csv(out_file, data, label_names=None):
    writer = csv.writer(out_file)
    if label_names:
        writer.writerow(label_names)
    writer.writerows(data)


def main():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('input_file', nargs='?')
    parser.add_argument('-o', '--output', help='File to write to (.json, .csv, .pkl, or .npy)')
    parser.add_argument('-t', '--type', help='Output format (json or csv)', choices=['json', 'csv'], default='csv')
    parser.add_argument('-n', '--no-header', action='store_true', help='Ignore label header')
    args = parser.parse_args()

    if not args.input_file and sys.stdin.isatty():
        parser.error('Please provide an input file or pipe input from stdin')

    if args.output:
        _, ext = splitext(args.output)
        if ext not in ['.json', '.csv', '.pkl', '.npy']:
            parser.error('Unknown output extension: ' + ext)

    if args.input_file:
        with open(args.input_file) as f:
            input_string = f.read()
    else:
        input_string = sys.stdin.read()

    result = parse(input_string)

    if args.no_header:
        result.labels.clear()

    if args.output:
        _, ext = splitext(args.output)
        if ext == '.json':
            with open(args.output, 'w') as f:
                json.dump(to_json(result.rows, result.labels), f)
        elif ext == '.csv':
            with open(args.output, 'w', newline='') as f:
                write_csv(f, result.rows, result.labels)
        elif ext == '.pkl':
            if result.labels:
                result.rows.insert(0, result.labels)
            with open(args.output, 'wb') as f:
                pickle.dump(result.rows, f)
        elif ext == '.npy':
            import numpy as np
            with open(args.output, 'wb') as f:
                np.save(f, np.array(result.rows))
    else:
        if args.type == 'json':
            print(json.dumps(to_json(result.rows, result.labels), indent=4 if sys.stdout.isatty() else None))
        elif args.type == 'csv':
            s = io.StringIO()
            write_csv(s, result.rows, result.labels)
            print(s.getvalue().strip())


if __name__ == '__main__':
    main()



import csv
import json
from os.path import isfile, splitext
from urllib import request

from avalares.parser import parse


def to_numpy(param):
    import numpy as np
    if param.startswith('http'):
        text = request.urlopen(param).read().decode()
    elif len(param) < 256 and isfile(param):
        _, ext = splitext(param)
        if ext == '.csv':
            with open(param, newline='') as csvfile:
                dialect = csv.Sniffer().sniff(csvfile.read(1024))
                csvfile.seek(0)
                reader = csv.reader(csvfile, dialect)
                return np.array(reader)
        elif ext == '.json':
            with open(param) as f:
                rows = json.load(f)
            if not isinstance(rows, list):
                raise ValueError('Cannot convert given json to numpy array')
            if rows and isinstance(rows[0], dict):
                labels = sorted(rows[0])
                return np.array([tuple(labels)] + [
                    tuple(row.get(k) for k in labels)
                    for row in rows
                ])
            else:
                return np.array(rows)
        else:
            with open(param) as f:
                text = f.read()
    else:
        text = param
    result = parse(text)
    return np.array(result.rows)


def to_pandas(param):
    import pandas as pd
    if param.startswith('http'):
        text = request.urlopen(param).read().decode()
    elif len(param) < 256 and isfile(param):
        _, ext = splitext(param)
        if ext == '.csv':
            return pd.read_csv(param)
        elif ext == '.json':
            return pd.read_json(param)
        else:
            with open(param) as f:
                text = f.read()
    else:
        text = param
    result = parse(text)
    return pd.DataFrame(result.rows, columns=result.labels)



from setuptools import setup

setup(
    name='avalares',
    version='0.1.1',
    description='Automatic text data extraction tool',
    url='https://github.com/matthewscholefield/avalares',
    author='Matthew D. Scholefield',
    author_email='matthew331199@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='avalares',
    packages=['avalares'],
    install_requires=[],
    entry_points={
        'console_scripts': [
            'avalares=avalares.__main__:main'
        ],
    }
)
